#include <e.h>
#include "e_mod_main.h"
#include <libmpd/libmpd.h>

/* Func Proto Requirements for Gadcon */
static E_Gadcon_Client *_gc_init     (E_Gadcon *gc, const char *name, const char *id, const char *style);
static void             _gc_shutdown (E_Gadcon_Client *gcc);
static void             _gc_orient   (E_Gadcon_Client *gcc);
static char            *_gc_label    (void);
static Evas_Object     *_gc_icon     (Evas *evas);

/* Module Protos */
static int           _isound_try_connect(void *data);
static Ecore_Timer*  _isound_mpd_init();
static void          _isound_mpd_signal_connect_status_changed(void *data);
static int           _isound_mpd_update(void *data);
static void          _isound_mpd_connection_changed(MpdObj *mo, int connect, void *data);
static void          _isound_parse_format(char *buf, size_t max_size, const char *format);

/* Callbacks Proto */
static void  _button_cb_mouse_down(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void  _menu_cb_post(void *data, E_Menu *m);
static void  _isound_status_changed_callback(MpdObj *mo, ChangedStatusType what, void *data);
static void  _isound_face_cb_menu_configure(void *data, E_Menu *m, E_Menu_Item *mi);
static void  _isound_cb_play (void *data, Evas_Object *obj, const char *emission, const char *source);
static void  _isound_cb_prev (void *data, Evas_Object *obj, const char *emission, const char *source);
static void  _isound_cb_next (void *data, Evas_Object *obj, const char *emission, const char *source);
static void  _isound_cb_stop (void *data, Evas_Object *obj, const char *emission, const char *source);
static void  _isound_cb_pause(void *data, Evas_Object *obj, const char *emission, const char *source);

/* Globals */
static E_Config_DD *conf_edd = NULL;
static Ecore_Timer *check_timer;
static MpdObj *mo;

Config *isound_config = NULL;

/******************************************************************************
 *                                                                            *
 *                           Gadcon Functions                                 *
 *                                                                            *
 * ****************************************************************************/
static const E_Gadcon_Client_Class _gadcon_class = 
{
  GADCON_CLIENT_CLASS_VERSION,
  "isound", 
  {
    _gc_init, _gc_shutdown, _gc_orient, _gc_label, _gc_icon
  },
  E_GADCON_CLIENT_STYLE_PLAIN
};


typedef struct _Instance Instance;
struct _Instance
{
  E_Gadcon_Client *gcc;
  Evas_Object *isound;
};

static Instance *inst;

static E_Gadcon_Client *
_gc_init (E_Gadcon * gc, const char *name, const char *id, const char *style)
{
  Evas_Object *o;
  E_Gadcon_Client *gcc;
  char buf[4096];

  inst = E_NEW (Instance, 1);

  o = edje_object_add(gc->evas);
  snprintf(buf, sizeof(buf), "%s/isound.edj", e_module_dir_get(isound_config->module));
  edje_object_file_set(o,buf, "isound");
  evas_object_show(o);

  gcc = e_gadcon_client_new (gc, name, id, style, o);
  gcc->data = inst;
  inst->gcc = gcc;
  inst->isound = o;

  evas_object_event_callback_add(o, EVAS_CALLBACK_MOUSE_DOWN, _button_cb_mouse_down, inst);
  edje_object_signal_callback_add(o, "play",  "isound", _isound_cb_play,  inst);
  edje_object_signal_callback_add(o, "prev",  "isound", _isound_cb_prev,  inst);
  edje_object_signal_callback_add(o, "next",  "isound", _isound_cb_next,  inst);
  edje_object_signal_callback_add(o, "stop",  "isound", _isound_cb_stop,  inst);
  edje_object_signal_callback_add(o, "pause", "isound", _isound_cb_pause, inst);

  check_timer = ecore_timer_add(0.2, _isound_try_connect, NULL);

  return gcc;
}

static void
_gc_shutdown (E_Gadcon_Client * gcc)
{
  inst = gcc->data;

  evas_object_event_callback_del(inst->isound, EVAS_CALLBACK_MOUSE_DOWN, _button_cb_mouse_down);
  evas_object_del(inst->isound);

  if (check_timer) ecore_timer_del(check_timer);
  check_timer = NULL;

  free (inst);
  inst = NULL;
}

static void
_gc_orient (E_Gadcon_Client * gcc)
{
  Evas_Coord mw, mh;

  inst = gcc->data;
  e_gadcon_client_min_size_set(gcc, 16, 16);
}

static char *
_gc_label (void)
{
  return _("ISound");
}

static Evas_Object *
_gc_icon (Evas * evas)
{
  Evas_Object *o;
  char buf[4096];

  o = edje_object_add (evas);
  snprintf(buf, sizeof (buf), "%s/module.edj", e_module_dir_get (isound_config->module));
  edje_object_file_set (o, buf, "icon");
  return o;
}

/******************************************************************************
 *                                                                            *
 *                           Isound Internals                                 *
 *                                                                            *
 * ****************************************************************************/
static int 
_isound_try_connect(void *data)
{
  Ecore_Timer *timer;

  if (!isound_config) 
    return 1;

  edje_object_message_signal_process(inst->isound);

  timer = _isound_mpd_init();

  if (!timer)
    {
      return 1;
    }
  else 
    {
      ecore_timer_del(check_timer);
      check_timer = timer;

      _isound_mpd_signal_connect_status_changed(NULL);
    }
}

Ecore_Timer*
_isound_mpd_init()
{
  Ecore_Timer *timer;
  const char *passwd = NULL;

  if (isound_config->passwd && isound_config->passwd[0] != '\0') 
    passwd = isound_config->passwd;

  mo = mpd_new((char*)isound_config->hostname, isound_config->port, (char*)passwd);
  if (!mpd_connect(mo))
    {
      if (passwd) mpd_send_password(mo);
      timer = ecore_timer_add(isound_config->poll_time, _isound_mpd_update, NULL);
    }
  else
    {
      timer = NULL;
    }

  return timer;
}

void
_isound_mpd_signal_connect_status_changed(void *data)
{
  mpd_signal_connect_status_changed(mo,
                                    (StatusChangedCallback)
                                    _isound_status_changed_callback, data);
  mpd_signal_connect_connection_changed(mo,
                                        (ConnectionChangedCallback)
                                        _isound_mpd_connection_changed, data);
}

int
_isound_mpd_update(void *data)
{
  mpd_status_update(mo);
  return 1;
}

static void
_isound_mpd_connection_changed(MpdObj *mo, int connect, void *data)
{
  if (!connect)
    {
      ecore_timer_del(check_timer);
      check_timer = ecore_timer_add(0.2, _isound_try_connect, NULL);
    }
}

void
_isound_config_updated(void)
{
  ecore_timer_del(check_timer);
  check_timer = ecore_timer_add(0.2, _isound_try_connect, NULL);
}

static void
_isound_parse_format(char *buf, size_t max_size, const char *format)
{
  mpd_Song *song = NULL;
  int elapsed_time, total_time;
  char artist[256], album[256], time[256], title[256];
  int i = 0, size = 0;
  const char *p = format;

  buf[0] = '\0';
  song = mpd_playlist_get_current_song(mo);
  if(!song) return;

  elapsed_time = mpd_status_get_elapsed_song_time(mo);
  total_time = mpd_status_get_total_song_time(mo);

  snprintf(artist, sizeof(artist), "%s", song->artist ? song->artist : "");
  snprintf(album,  sizeof(album),  "%s", song->album  ? song->album  : "");
  snprintf(title,  sizeof(title),  "%s", song->title  ? song->title  : "");
  snprintf(time,   sizeof(time),   "%d:%02d / %d:%02d", 
           elapsed_time / 60, elapsed_time % 60, 
           total_time / 60, total_time % 60);

  while (format && *p != '\0' && size < max_size)
    {
      if (*p == '%')
        {
          p++;
          if (*p == 'a')
            {
              size += strlen(artist);
              if (size >= max_size) break;
              sprintf(buf, "%s%s", buf, artist);
            }
          else if (*p == 'b')
            {
              size += strlen(album);
              if (size >= max_size) break;
              sprintf(buf, "%s%s", buf, album);
            }
          else if (*p == 't')
            {
              size += strlen(title);
              if (size >= max_size) break;
              sprintf(buf, "%s%s", buf, title);
            }
          else if (*p == 'm')
            {
              size += strlen(time);
              if (size >= max_size) break;
              sprintf(buf, "%s%s", buf, time);
            }
          else
            {
              buf[size] = '%';
              buf[size+1] = *p;
              buf[size+2] = '\0';
              size += 2;
            }
        }
      else
        {
          buf[size] = *p;
          buf[size+1] = '\0';
          size++;
        }
      p++;
    }
}

/******************************************************************************
 *                                                                            *
 *                           Isound Callbacks                                 *
 *                                                                            *
 * ****************************************************************************/
static void
_button_cb_mouse_down(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
  Evas_Event_Mouse_Down *ev;

  inst = data;
  ev = event_info;

  if ((ev->button == 3) && (!isound_config->menu))
    {
      E_Menu *mn;
      E_Menu_Item *mi;
      int cx, cy, cw, ch;

      mn = e_menu_new();
      e_menu_post_deactivate_callback_set(mn, _menu_cb_post, inst);
      isound_config->menu = mn;

      mi = e_menu_item_new(mn);
      e_menu_item_label_set(mi, _("Configuration"));
      e_util_menu_item_edje_icon_set(mi, "enlightenment/configuration");   
      e_menu_item_callback_set(mi, _isound_face_cb_menu_configure, NULL);

      e_gadcon_client_util_menu_items_append(inst->gcc, mn, 0);

      e_gadcon_canvas_zone_geometry_get(inst->gcc->gadcon,
                                        &cx, &cy, &cw, &ch);
      e_menu_activate_mouse(mn,
                            e_util_zone_current_get(e_manager_current_get()),
                            cx + ev->output.x, cy + ev->output.y, 1, 1,
                            E_MENU_POP_DIRECTION_DOWN, ev->timestamp);
      e_util_evas_fake_mouse_up_later(inst->gcc->gadcon->evas,
                                      ev->button);
    }
}

static void
_menu_cb_post(void *data, E_Menu *m)
{
  if (!isound_config->menu) return;
  e_object_del(E_OBJECT(isound_config->menu));
  isound_config->menu = NULL;
}

void
_isound_status_changed_callback(MpdObj *blabla, ChangedStatusType what, void *data)
{
  mpd_Song *song = NULL;
  MpdState state;
  int elapsed_time, total_time;
  char buf[4096];

  if (what & MPD_CST_ELAPSED_TIME)
    {
      if (isound_config->line1_format 
          && strstr(isound_config->line1_format, "%m"))
        {
          _isound_parse_format(buf, 4096, isound_config->line1_format);
          edje_object_part_text_set(inst->isound, "line1", buf);
        }
      if (isound_config->line2_format 
          && strstr(isound_config->line2_format, "%m"))
        {
          _isound_parse_format(buf, 4096, isound_config->line2_format);
          edje_object_part_text_set(inst->isound, "line2", buf);
        }
      if (isound_config->line3_format 
          && strstr(isound_config->line3_format, "%m"))
        {
          _isound_parse_format(buf, 4096, isound_config->line1_format);
          edje_object_part_text_set(inst->isound, "line3", buf);
        }
    }
  if (what & MPD_CST_STATE)
    {
      state = mpd_player_get_state(mo);
      switch (state)
        {
        case MPD_PLAYER_STOP:
          //TODO
          snprintf(buf, sizeof(buf), "Player stopped");
          edje_object_part_text_set(inst->isound, "line1", buf);
          edje_object_part_text_set(inst->isound, "line2", "");
          edje_object_part_text_set(inst->isound, "line3", "");
          edje_object_signal_emit(inst->isound, "mode,single", "isound");
          break;
        case MPD_PLAYER_PAUSE:
          //TODO
          snprintf(buf, sizeof(buf), "Player paused");
          edje_object_part_text_set(inst->isound, "line1", buf);
          edje_object_part_text_set(inst->isound, "line2", "");
          edje_object_part_text_set(inst->isound, "line3", "");
          edje_object_signal_emit(inst->isound, "mode,single", "isound");
          break;
        case MPD_PLAYER_PLAY:
          //TODO
          what |= MPD_CST_SONGID;
          break;
        case MPD_PLAYER_UNKNOWN:
          snprintf(buf, sizeof(buf), "Unknown status");
          edje_object_part_text_set(inst->isound, "line1", buf);
          break;
        }
    }
  if (what & MPD_CST_SONGID)
    {
      song = mpd_playlist_get_current_song(mo);
      if(song) 
        {
          int mode = 0;
          char line[6];

          edje_object_part_text_set(inst->isound, "line1", "");
          edje_object_part_text_set(inst->isound, "line2", "");
          edje_object_part_text_set(inst->isound, "line3", "");

          if (isound_config->line1_format && isound_config->line1_format[0] != '\0')
            {
              mode++;
              _isound_parse_format(buf, 4096, isound_config->line1_format);
              snprintf(line, sizeof(line), "line%d", mode);
              edje_object_part_text_set(inst->isound, line, buf);
            }
          if (isound_config->line2_format && isound_config->line2_format[0] != '\0')
            {
              mode++;
              _isound_parse_format(buf, 4096, isound_config->line2_format);
              snprintf(line, sizeof(line), "line%d", mode);
              edje_object_part_text_set(inst->isound, line, buf);
            }
          if (isound_config->line3_format && isound_config->line3_format[0] != '\0')
            {
              mode++;
              _isound_parse_format(buf, 4096, isound_config->line3_format);
              snprintf(line, sizeof(line), "line%d", mode);
              printf("%s\n", line);
              edje_object_part_text_set(inst->isound, line, buf);
            }

          if (mode == 1)
            edje_object_signal_emit(inst->isound, "mode,single", "isound");
          else if (mode == 2)
            edje_object_signal_emit(inst->isound, "mode,double", "isound");
          else 
            edje_object_signal_emit(inst->isound, "mode,triple", "isound");

          if (isound_config->resize)
            {
              int w, h, ow;
              Evas_Object *o;

              edje_object_calc_force(inst->isound);

              o = edje_object_part_object_get(inst->isound, "line1");
              evas_object_textblock_size_formatted_get(o, &ow, NULL);
              w = ow;
              o = edje_object_part_object_get(inst->isound, "line2");
              evas_object_textblock_size_formatted_get(o, &ow, NULL);
              if (ow > w) w = ow;
              o = edje_object_part_object_get(inst->isound, "line3");
              evas_object_textblock_size_formatted_get(o, &ow, NULL);
              if (ow > w) w = ow;
              evas_object_geometry_get(inst->isound, NULL, NULL, NULL, &h);
              e_gadcon_client_aspect_set(inst->gcc, w, h);
              evas_object_resize(inst->isound, w, h);
            }
        }
    }
}

static void 
_isound_face_cb_menu_configure(void *data, E_Menu *m, E_Menu_Item *mi) 
{
  if (!isound_config) return;
  if (isound_config->config_dialog) return;
  _config_isound_module();
}


static void
_isound_cb_play(void *data, Evas_Object *obj, const char *emission, const char *source)
{
  mpd_player_play(mo);
}

static void
_isound_cb_prev(void *data, Evas_Object *obj, const char *emission, const char *source)
{
  mpd_player_prev(mo);
}

static void
_isound_cb_next(void *data, Evas_Object *obj, const char *emission, const char *source)
{
  mpd_player_next(mo);
}

static void
_isound_cb_stop(void *data, Evas_Object *obj, const char *emission, const char *source)
{
  mpd_player_stop(mo);
}

static void
_isound_cb_pause(void *data, Evas_Object *obj, const char *emission, const char *source)
{
  mpd_player_pause(mo);
}

/******************************************************************************
 *                                                                            *
 *                                 Module API                                 *
 *                                                                            *
 * ****************************************************************************/
EAPI E_Module_Api e_modapi = {
  E_MODULE_API_VERSION,
  "ISound"
};

EAPI void *
e_modapi_init (E_Module * m)
{

  conf_edd = E_CONFIG_DD_NEW("Isound_Config", Config);
#undef T
#undef D
#define T Config
#define D conf_edd
  E_CONFIG_VAL(D, T, poll_time, DOUBLE);
  E_CONFIG_VAL(D, T, resize, INT);
  E_CONFIG_VAL(D, T, line1_format, STR);
  E_CONFIG_VAL(D, T, line2_format, STR);
  E_CONFIG_VAL(D, T, line3_format, STR);
  E_CONFIG_VAL(D, T, hostname, STR);
  E_CONFIG_VAL(D, T, port, INT);
  E_CONFIG_VAL(D, T, passwd, STR);

  isound_config = e_config_domain_load("module.isound", conf_edd);  
  if (!isound_config)
    {
      isound_config = E_NEW(Config, 1);
      isound_config->resize = 0; 
      isound_config->poll_time = 0.5;
      isound_config->line1_format = evas_stringshare_add("%t / <b>%a</b>");
      isound_config->line2_format = evas_stringshare_add("<font_size=9>%b  [%m]");
      isound_config->line3_format = NULL;
      isound_config->hostname = evas_stringshare_add("localhost");
      isound_config->port = 6600;
      isound_config->passwd = evas_stringshare_add("");
    }

  E_CONFIG_LIMIT(isound_config->resize, 0, 1);
  E_CONFIG_LIMIT(isound_config->poll_time, 0.2, 5.0);

  isound_config->module = m;

  e_gadcon_provider_register(&_gadcon_class);

  return m;
}

EAPI int
e_modapi_shutdown (E_Module * m)
{
  isound_config->module = NULL;
  e_gadcon_provider_unregister (&_gadcon_class);

  if (isound_config->line1_format)
    evas_stringshare_del(isound_config->line1_format);
  if (isound_config->line2_format)
    evas_stringshare_del(isound_config->line2_format);
  if (isound_config->line3_format)
    evas_stringshare_del(isound_config->line3_format);
  if (isound_config->hostname)
    evas_stringshare_del(isound_config->hostname);
  if (isound_config->passwd)
    evas_stringshare_del(isound_config->passwd);

  free(isound_config);
  isound_config = NULL;
  E_CONFIG_DD_FREE(conf_edd);

  return 1;
}

EAPI int
e_modapi_save (E_Module * m)
{
  e_config_domain_save("module.isound", conf_edd, isound_config);
  return 1;
}

EAPI int
e_modapi_about (E_Module * m)
{
  e_module_dialog_show (m, _("ISound module"),
                        _("A module to display informations of the current playing song"));
  return 1;
}
